/*
PeRdr - PE file disassembler
Copyright (C) 1999-2002 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#include "global.h"
#ifdef HAVE_HDRSTOP
#pragma hdrstop
#endif

#include "bytelevel.h"


// sezione errata. non contiene ne dati ne codice
const FileByte::Section FileByte::errorSection;

FileByte::FileByte(CFile& file,const RVAFileTranslator& rva,uint32_t imageBase,
                   bool _hasRelocation,const RelocationInfos& _relocationInfos):
  sectionBuffer(NULL),
  sections(rva.GetSectionCount()),currSection(0),
  currPos(0),
  hasRelocation(_hasRelocation),relocationInfos(_relocationInfos)
{
  unsigned numSection = rva.GetSectionCount();
  sectionBuffer = new uint8_t*[numSection];
  for (unsigned n=0;n<numSection;++n)
  {
    // se RawSize > RvaSize altri bytes non utilizzati (indirizzi invalidi)
    // se RvaSize > RawSize altri bytes non inizializzati
    // calcola dimensione sezione
	  unsigned size = rva.GetSection(n)->RvaSize;
//		  rva.GetSection(n)->RvaSize > rva.GetSection(n)->RawSize ?
//		  rva.GetSection(n)->RvaSize : rva.GetSection(n)->RawSize;
    // inizializza valori
    sections[n].begin = rva.GetSection(n)->RvaAddress + imageBase;
    sections[n].end   = rva.GetSection(n)->RvaAddress + imageBase + size;
    if ( rva.GetSection(n)->IsCode() )
      sections[n].SetIsCode();
    else
      sections[n].SetIsData();

    // inizializza buffer
    sectionBuffer[n] = new uint8_t[size];
    memset(sectionBuffer[n],0,size);
    // se puntatore valido leggi da memoria
    if ( rva.GetSection(n)->RawAddress != 0 )
    {
      unsigned rawSize = rva.GetSection(n)->RawSize;
      file.Seek( rva.GetSection(n)->RawAddress );
//      file.RawRead( sectionBuffer[n],rva.GetSection(n)->RawSize );
      file.RawRead( sectionBuffer[n],size<rawSize?size:rawSize );
    }
  }
  currPos = 0;
  currPos = GetNextValid();
  SeekTo(currPos);
  _PRG_ASSERT(currPos!=0 && currSection!=0);
}

FileByte::~FileByte()
{
  for( TSections::size_type n=0; n< sections.size(); ++n)
    delete[] sectionBuffer[n];
  delete[] sectionBuffer;
}

uint8_t FileByte::ReadByte()
{
  // test validita' cella
  if (!currSection || currPos>=currSection->end)
  {
//#ifdef DEBUG
//    printf("mao");
//#endif
//  	throw OutOfAddress(currPos);
    // !!! overlapped section !!!
    SeekTo(currPos);
  }
  _PRG_ASSERT(IsValid(currPos));

  uint8_t u;
  u = *pSectionBuffer;
  ++pSectionBuffer;

  // !!! questo deve essere fatto dopo la lettura per problemi di eccezione
  ++currPos;
  return u;
}

class FileByteSavePos
{
public:
  FileByteSavePos(FileByte& _fb,vma_t addr):fb(_fb),address(addr) {};
  ~FileByteSavePos() { fb.SeekTo(address); }
private:
  FileByte& fb;
  vma_t address;
};

uint8_t FileByte::ReadByte(vma_t address)
{
  FileByteSavePos save(*this,currPos);
  SeekTo(address);
  return ReadByte();
}

uint16_t FileByte::ReadWord()
{
  // !!! define uint8fast_t
  uint8_t u = ReadByte();
  return uint16_t(u|(uint16_t(ReadByte())<<8));
}

uint16_t FileByte::ReadWord(vma_t address)
{
  FileByteSavePos save(*this,currPos);
  SeekTo(address);
  return ReadWord();
}

uint32_t FileByte::ReadDword()
{
  // !!! define uint16fast_t
  uint16_t u = ReadWord();
  return u|(uint32_t(ReadWord())<<16);
}

uint32_t FileByte::ReadDword(vma_t address)
{
  FileByteSavePos save(*this,currPos);
  SeekTo(address);
  return ReadDword();
}

void FileByte::SeekTo(vma_t address)
{
  // prova con la sezione corrente
  if ( currSection && currSection->begin <= address && address < currSection->end )
  {
    pSectionBuffer = pSectionBuffer - currPos + address;
    currPos = address;
    return;
  }

  // cerca una sezione
  for( TSections::size_type n=0; n < sections.size(); ++n )
  {
    vma_t begin = sections[n].begin;
    vma_t end   = sections[n].end;
    if (begin <= address && address < end)
    {
      currSection = &sections[n];
      pSectionBuffer = sectionBuffer[n] + (address-begin);
      currPos = address;
      return;
    }
  }

  throw OutOfAddress(address);
}

vma_t FileByte::Tell()
{
  return currPos;
}

// implementazione simile a SeekTo
bool FileByte::IsValid(vma_t address)
{
  // prova con la sezione corrente
  if ( currSection && currSection->begin <= address && address < currSection->end )
    return true;

  // cerca sezione
  for( TSections::size_type n=0; n < sections.size(); ++n )
  {
    vma_t begin = sections[n].begin;
    vma_t end   = sections[n].end;
    if (begin <= address && address < end)
    {
      return true;
    }
  }
	return false;
}

vma_t FileByte::GetNextValid()
{
  if (IsValid(currPos))
    return currPos;

  vma_t pos = VMA_MAX;

  // trova il minimo fra le sezioni maggiori
  for( TSections::size_type n=0; n < sections.size(); ++n )
  {
    vma_t begin = sections[n].begin;
    if (begin < pos && begin >= currPos)
      pos = begin;
  }
  _PRG_ASSERT(IsValid(pos));
  return pos;
}

const FileByte::Section* FileByte::GetSection(vma_t address) const
{
  // cerca sezione
  for( TSections::size_type n=0; n < sections.size(); ++n )
  {
    vma_t begin = sections[n].begin;
    vma_t end   = sections[n].end;
    if (begin <= address && address < end)
    {
      return &sections[n];
    }
  }
  return &errorSection;
}

